@@ -396,6 +396,11 @@ func validateAlertmanagerConfig(cfg interface{}) error {
 			return err
 		}
 
+	case reflect.TypeOf(config.MSTeamsV2Config{}):
+		if err := validateMSTeamsV2Config(v.Interface().(config.MSTeamsV2Config)); err != nil {
+			return err
+		}
+
 	case reflect.TypeOf(config.TelegramConfig{}):
 		if err := validateTelegramConfig(v.Interface().(config.TelegramConfig)); err != nil {
 			return err
@@ -586,6 +591,15 @@ func validateMSTeamsConfig(cfg config.MSTeamsConfig) error {
 	return nil
 }
 
+// validateMSTeamsV2Config validates the Microsoft Teams config and returns an error if it
+// contains settings not allowed by Mimir.
+func validateMSTeamsV2Config(cfg config.MSTeamsV2Config) error {
+	if cfg.WebhookURLFile != "" {
+		return errWebhookURLFileNotAllowed
+	}
+	return nil
+}
+
 // validateTelegramConfig validates the Telegram config and returns an error if it contains
 // settings not allowed by Mimir.
 func validateTelegramConfig(cfg config.TelegramConfig) error {
