@@ -4,6 +4,7 @@ package blockbuilder
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"os"
 	"path/filepath"
@@ -14,19 +15,17 @@ import (
 	"github.com/go-kit/log"
 	"github.com/go-kit/log/level"
 	"github.com/grafana/dskit/multierror"
+	dskittenant "github.com/grafana/dskit/tenant"
 	"github.com/oklog/ulid/v2"
-	"github.com/pkg/errors"
 	"github.com/prometheus/prometheus/model/histogram"
 	"github.com/prometheus/prometheus/model/labels"
 	"github.com/prometheus/prometheus/storage"
 	"github.com/prometheus/prometheus/tsdb"
 	"github.com/prometheus/prometheus/tsdb/chunks"
-	"github.com/twmb/franz-go/pkg/kgo"
 	"golang.org/x/sync/errgroup"
 
 	"github.com/grafana/mimir/pkg/mimirpb"
 	mimir_storage "github.com/grafana/mimir/pkg/storage"
-	"github.com/grafana/mimir/pkg/storage/ingest"
 	mimir_tsdb "github.com/grafana/mimir/pkg/storage/tsdb"
 	util_log "github.com/grafana/mimir/pkg/util/log"
 	"github.com/grafana/mimir/pkg/util/validation"
@@ -41,6 +40,8 @@ type TSDBBuilder struct {
 	metrics                          tsdbBuilderMetrics
 	applyMaxGlobalSeriesPerUserBelow int // inclusive
 
+	partitionID int32
+
 	// Map of a tenant in a partition to its TSDB.
 	tsdbsMu sync.RWMutex
 	tsdbs   map[tsdbTenant]*userTSDB
@@ -62,7 +63,7 @@ type tsdbTenant struct {
 	tenantID    string
 }
 
-func NewTSDBBuilder(logger log.Logger, dataDir string, blocksStorageCfg mimir_tsdb.BlocksStorageConfig, limits *validation.Overrides, metrics tsdbBuilderMetrics, applyMaxGlobalSeriesPerUserBelow int) *TSDBBuilder {
+func NewTSDBBuilder(logger log.Logger, dataDir string, partitionID int32, blocksStorageCfg mimir_tsdb.BlocksStorageConfig, limits *validation.Overrides, metrics tsdbBuilderMetrics, applyMaxGlobalSeriesPerUserBelow int) *TSDBBuilder {
 	return &TSDBBuilder{
 		dataDir:                          dataDir,
 		logger:                           logger,
@@ -70,27 +71,19 @@ func NewTSDBBuilder(logger log.Logger, dataDir string, blocksStorageCfg mimir_ts
 		blocksStorageCfg:                 blocksStorageCfg,
 		metrics:                          metrics,
 		applyMaxGlobalSeriesPerUserBelow: applyMaxGlobalSeriesPerUserBelow,
+		partitionID:                      partitionID,
 		tsdbs:                            make(map[tsdbTenant]*userTSDB),
 	}
 }
 
-// Process puts the samples in the TSDB. Some parts taken from (*Ingester).pushSamplesToAppender.
-func (b *TSDBBuilder) Process(ctx context.Context, rec *kgo.Record) (err error) {
-	userID := string(rec.Key)
-
-	req := mimirpb.PreallocWriteRequest{
-		SkipUnmarshalingExemplars: true,
-	}
+// PushToStorageAndReleaseRequest implements ingest.Pusher.
+// It puts the samples in the TSDB. Some parts taken from (*Ingester).pushSamplesToAppender.
+func (b *TSDBBuilder) PushToStorageAndReleaseRequest(ctx context.Context, req *mimirpb.WriteRequest) error {
 	defer mimirpb.ReuseSlice(req.Timeseries)
 
-	version := ingest.ParseRecordVersion(rec)
-	if version > ingest.LatestRecordVersion {
-		return fmt.Errorf("received a record with an unsupported version: %d, max supported version: %d", version, ingest.LatestRecordVersion)
-	}
-
-	err = ingest.DeserializeRecordContent(rec.Value, &req, version)
+	tenantID, err := dskittenant.TenantID(ctx)
 	if err != nil {
-		return fmt.Errorf("unmarshal record key %s: %w", rec.Key, err)
+		return fmt.Errorf("extract tenant id: %w", err)
 	}
 
 	if len(req.Timeseries) == 0 {
@@ -98,26 +91,26 @@ func (b *TSDBBuilder) Process(ctx context.Context, rec *kgo.Record) (err error)
 	}
 
 	tenant := tsdbTenant{
-		partitionID: rec.Partition,
-		tenantID:    userID,
+		partitionID: b.partitionID,
+		tenantID:    tenantID,
 	}
 	db, err := b.getOrCreateTSDB(tenant)
 	if err != nil {
-		return fmt.Errorf("get tsdb for tenant %s: %w", userID, err)
+		return fmt.Errorf("get tsdb for tenant %s: %w", tenantID, err)
 	}
 
 	app := db.Appender(ctx).(extendedAppender)
 	defer func() {
 		if err != nil {
 			if e := app.Rollback(); e != nil && !errors.Is(e, tsdb.ErrAppenderClosed) {
-				level.Warn(b.logger).Log("msg", "failed to rollback appender on error", "tenant", userID, "err", e)
+				level.Warn(b.logger).Log("msg", "failed to rollback appender on error", "tenant", tenantID, "err", e)
 			}
 			// Always wrap the returned error with tenant.
-			err = fmt.Errorf("failed to process record for tenant %s: %w", userID, err)
+			err = fmt.Errorf("failed to process record for tenant %s: %w", tenantID, err)
 		}
 	}()
 
-	nativeHistogramsIngestionEnabled := b.limits.NativeHistogramsIngestionEnabled(userID)
+	nativeHistogramsIngestionEnabled := b.limits.NativeHistogramsIngestionEnabled(tenantID)
 
 	var (
 		labelsBuilder   labels.ScratchBuilder
@@ -145,8 +138,14 @@ func (b *TSDBBuilder) Process(ctx context.Context, rec *kgo.Record) (err error)
 					copiedLabels = mimirpb.CopyLabels(nonCopiedLabels)
 					ref, err = app.AppendCTZeroSample(0, copiedLabels, s.TimestampMs, ts.CreatedTimestamp)
 				}
-				if err != nil && !errors.Is(err, storage.ErrOutOfOrderCT) && !errors.Is(err, storage.ErrOutOfOrderSample) {
-					level.Warn(b.logger).Log("msg", "failed to store zero float sample for created timestamp", "tenant", userID, "err", err)
+				if err != nil && !errors.Is(err, storage.ErrDuplicateSampleForTimestamp) && !errors.Is(err, storage.ErrOutOfOrderCT) && !errors.Is(err, storage.ErrOutOfOrderSample) {
+					// According to OTEL spec: https://opentelemetry.io/docs/specs/otel/metrics/data-model/#cumulative-streams-handling-unknown-start-time
+					// if the start time is unknown, then it should equal to the timestamp of the first sample,
+					// which will mean a created timestamp equal to the timestamp of the first sample for later
+					// samples. Thus we ignore if zero sample would cause duplicate.
+					// We also ignore out of order sample as created timestamp is out of order most of the time,
+					// except when written before the first sample.
+					level.Warn(b.logger).Log("msg", "failed to store zero float sample for created timestamp", "tenant", tenantID, "err", err)
 					discardedSamples++
 				}
 				ingestCreatedTimestamp = false // Only try to append created timestamp once per series.
@@ -199,8 +198,14 @@ func (b *TSDBBuilder) Process(ctx context.Context, rec *kgo.Record) (err error)
 					copiedLabels = mimirpb.CopyLabels(nonCopiedLabels)
 					ref, err = app.AppendHistogramCTZeroSample(0, copiedLabels, h.Timestamp, ts.CreatedTimestamp, ih, fh)
 				}
-				if err != nil && !errors.Is(err, storage.ErrOutOfOrderCT) && !errors.Is(err, storage.ErrOutOfOrderSample) {
-					level.Warn(b.logger).Log("msg", "failed to store zero histogram sample for created timestamp", "tenant", userID, "err", err)
+				if err != nil && !errors.Is(err, storage.ErrDuplicateSampleForTimestamp) && !errors.Is(err, storage.ErrOutOfOrderCT) && !errors.Is(err, storage.ErrOutOfOrderSample) {
+					// According to OTEL spec: https://opentelemetry.io/docs/specs/otel/metrics/data-model/#cumulative-streams-handling-unknown-start-time
+					// if the start time is unknown, then it should equal to the timestamp of the first sample,
+					// which will mean a created timestamp equal to the timestamp of the first sample for later
+					// samples. Thus we ignore if zero sample would cause duplicate.
+					// We also ignore out of order sample as created timestamp is out of order most of the time,
+					// except when written before the first sample.
+					level.Warn(b.logger).Log("msg", "failed to store zero histogram sample for created timestamp", "tenant", tenantID, "err", err)
 					discardedSamples++
 				}
 				ingestCreatedTimestamp = false // Only try to append created timestamp once per series.
@@ -329,6 +334,7 @@ func (b *TSDBBuilder) newTSDB(tenant tsdbTenant) (*userTSDB, error) {
 		SeriesLifecycleCallback:              udb,
 		HeadPostingsForMatchersCacheMetrics:  tsdb.NewPostingsForMatchersCacheMetrics(nil),
 		BlockPostingsForMatchersCacheMetrics: tsdb.NewPostingsForMatchersCacheMetrics(nil),
+		PostingsClonerFactory:                tsdb.DefaultPostingsClonerFactory{},
 	}, nil)
 	if err != nil {
 		return nil, err
@@ -341,6 +347,10 @@ func (b *TSDBBuilder) newTSDB(tenant tsdbTenant) (*userTSDB, error) {
 	return udb, nil
 }
 
+func (b *TSDBBuilder) NotifyPreCommit(_ context.Context) error {
+	return nil
+}
+
 // Function to upload the blocks.
 type blockUploader func(_ context.Context, tenantID, dbDir string, metas []tsdb.BlockMeta) error
 
@@ -469,7 +479,7 @@ var (
 func (u *userTSDB) PreCreation(labels.Labels) error {
 	// Global series limit.
 	if u.maxGlobalSeries > 0 && u.Head().NumSeries() >= uint64(u.maxGlobalSeries) {
-		return errors.Wrapf(errMaxInMemorySeriesReached, "limit of %d reached for user %s", u.maxGlobalSeries, u.userID)
+		return fmt.Errorf("limit of %d reached for user %s: %w", u.maxGlobalSeries, u.userID, errMaxInMemorySeriesReached)
 	}
 
 	return nil
