@@ -7,6 +7,7 @@ package querymiddleware
 
 import (
 	"bytes"
+	"cmp"
 	"context"
 	"errors"
 	"fmt"
@@ -42,6 +43,7 @@ import (
 	"github.com/grafana/mimir/pkg/streamingpromql/compat"
 	"github.com/grafana/mimir/pkg/util"
 	"github.com/grafana/mimir/pkg/util/chunkinfologger"
+	"github.com/grafana/mimir/pkg/util/propagation"
 	"github.com/grafana/mimir/pkg/util/spanlogger"
 )
 
@@ -52,9 +54,9 @@ var (
 	allFormats        = []string{formatJSON, formatProtobuf}
 
 	// List of HTTP headers to propagate when a Prometheus request is encoded into a HTTP request.
-	// api.ReadConsistencyHeader is propagated as HTTP header -> Request.Context -> Request.Header, so there's no need to explicitly propagate it here.
+	// Read consistency level and max delay headers are propagated as HTTP header -> Request.Context -> Request.Header, so there's no need to explicitly propagate it here.
 	codecPropagateHeadersMetrics = []string{compat.ForceFallbackHeaderName, chunkinfologger.ChunkInfoLoggingHeader, api.ReadConsistencyOffsetsHeader, querier.FilterQueryablesHeader}
-	// api.ReadConsistencyHeader is propagated as HTTP header -> Request.Context -> Request.Header, so there's no need to explicitly propagate it here.
+	// Read consistency level and max delay headers are propagated as HTTP header -> Request.Context -> Request.Header, so there's no need to explicitly propagate it here.
 	codecPropagateHeadersLabels = []string{api.ReadConsistencyOffsetsHeader, querier.FilterQueryablesHeader}
 )
 
@@ -68,9 +70,6 @@ const (
 
 	totalShardsControlHeader = "Sharding-Control"
 
-	// Instant query specific options
-	instantSplitControlHeader = "Instant-Split-Control"
-
 	operationEncode = "encode"
 	operationDecode = "decode"
 
@@ -220,6 +219,7 @@ type Codec struct {
 	lookbackDelta                                   time.Duration
 	preferredQueryResultResponseFormat              string
 	propagateHeadersMetrics, propagateHeadersLabels []string
+	injector                                        propagation.Injector
 }
 
 type formatter interface {
@@ -237,7 +237,7 @@ var jsonFormatterInstance = jsonFormatter{}
 
 var knownFormats = []formatter{
 	jsonFormatterInstance,
-	protobufFormatter{},
+	ProtobufFormatter{},
 }
 
 func NewCodec(
@@ -245,6 +245,7 @@ func NewCodec(
 	lookbackDelta time.Duration,
 	queryResultResponseFormat string,
 	propagateHeaders []string,
+	injector propagation.Injector,
 ) Codec {
 	return Codec{
 		metrics:                            newCodecMetrics(registerer),
@@ -252,13 +253,14 @@ func NewCodec(
 		preferredQueryResultResponseFormat: queryResultResponseFormat,
 		propagateHeadersMetrics:            append(codecPropagateHeadersMetrics, propagateHeaders...),
 		propagateHeadersLabels:             append(codecPropagateHeadersLabels, propagateHeaders...),
+		injector:                           injector,
 	}
 }
 
 // MergeResponse merges responses from multiple requests into a single Response
 func (Codec) MergeResponse(responses ...Response) (Response, error) {
 	if len(responses) == 0 {
-		return newEmptyPrometheusResponse(), nil
+		return NewEmptyPrometheusResponse(), nil
 	}
 
 	promResponses := make([]*PrometheusResponse, 0, len(responses))
@@ -301,7 +303,17 @@ func (Codec) MergeResponse(responses ...Response) (Response, error) {
 	}
 
 	// Merge the responses.
-	sort.Sort(byFirstTime(promResponses))
+	slices.SortFunc(promResponses, func(a, b *PrometheusResponse) int {
+		aTime := int64(-1)
+		if len(a.Data.Result) > 0 && len(a.Data.Result[0].Samples) > 0 {
+			aTime = a.Data.Result[0].Samples[0].TimestampMs
+		}
+		bTime := int64(-1)
+		if len(b.Data.Result) > 0 && len(b.Data.Result[0].Samples) > 0 {
+			bTime = b.Data.Result[0].Samples[0].TimestampMs
+		}
+		return cmp.Compare(aTime, bTime)
+	})
 
 	return &PrometheusResponseWithFinalizer{
 		PrometheusResponse: &PrometheusResponse{
@@ -351,7 +363,7 @@ func (c Codec) decodeRangeQueryRequest(r *http.Request) (MetricsQueryRequest, er
 	}
 
 	var options Options
-	decodeOptions(r, &options)
+	DecodeOptions(r, &options)
 
 	stats := reqValues.Get("stats")
 	req := NewPrometheusRangeQueryRequest(
@@ -378,7 +390,7 @@ func (c Codec) decodeInstantQueryRequest(r *http.Request) (MetricsQueryRequest,
 	}
 
 	var options Options
-	decodeOptions(r, &options)
+	DecodeOptions(r, &options)
 
 	stats := reqValues.Get("stats")
 
@@ -396,7 +408,9 @@ func httpHeadersToProm(httpH http.Header) []*PrometheusHeader {
 	for h, hv := range httpH {
 		headers = append(headers, &PrometheusHeader{Name: h, Values: slices.Clone(hv)})
 	}
-	sort.Slice(headers, func(i, j int) bool { return headers[i].Name < headers[j].Name })
+	slices.SortFunc(headers, func(a, b *PrometheusHeader) int {
+		return strings.Compare(a.Name, b.Name)
+	})
 	return headers
 }
 
@@ -642,7 +656,7 @@ func decodeQueryMinMaxTime(queryExpr parser.Expr, start, end, step int64, lookba
 	return minTime, maxTime
 }
 
-func decodeOptions(r *http.Request, opts *Options) {
+func DecodeOptions(r *http.Request, opts *Options) {
 	opts.CacheDisabled = decodeCacheDisabledOption(r)
 
 	for _, value := range r.Header.Values(totalShardsControlHeader) {
@@ -655,18 +669,6 @@ func decodeOptions(r *http.Request, opts *Options) {
 			opts.ShardingDisabled = true
 		}
 	}
-
-	for _, value := range r.Header.Values(instantSplitControlHeader) {
-		splitInterval, err := time.ParseDuration(value)
-		if err != nil {
-			continue
-		}
-		// Instant split by time interval unit stored in nanoseconds (time.Duration unit in int64)
-		opts.InstantSplitInterval = splitInterval.Nanoseconds()
-		if opts.InstantSplitInterval < 1 {
-			opts.InstantSplitDisabled = true
-		}
-	}
 }
 
 func decodeCacheDisabledOption(r *http.Request) bool {
@@ -733,8 +735,22 @@ func (c Codec) EncodeMetricsQueryRequest(ctx context.Context, r MetricsQueryRequ
 		return nil, fmt.Errorf("unknown query result response format '%s'", c.preferredQueryResultResponseFormat)
 	}
 
-	if level, ok := api.ReadConsistencyLevelFromContext(ctx); ok {
-		req.Header.Add(api.ReadConsistencyHeader, level)
+	if err := c.AddHeadersForMetricQueryRequest(ctx, r, propagation.HttpHeaderCarrier(req.Header)); err != nil {
+		return nil, err
+	}
+
+	// Inject auth from context.
+	// This isn't included in AddHeadersForMetricQueryRequest as it's not needed for Protobuf requests to queriers.
+	if err := user.InjectOrgIDIntoHTTPRequest(ctx, req); err != nil {
+		return nil, err
+	}
+
+	return req.WithContext(ctx), nil
+}
+
+func (c Codec) AddHeadersForMetricQueryRequest(ctx context.Context, r MetricsQueryRequest, headers propagation.Carrier) error {
+	if err := c.injector.InjectToCarrier(ctx, headers); err != nil {
+		return nil
 	}
 
 	// Propagate allowed HTTP headers.
@@ -744,17 +760,11 @@ func (c Codec) EncodeMetricsQueryRequest(ctx context.Context, r MetricsQueryRequ
 		}
 
 		for _, v := range h.Values {
-			// There should only be one value, but add all of them for completeness.
-			req.Header.Add(h.Name, v)
+			headers.Add(h.Name, v)
 		}
 	}
 
-	// Inject auth from context.
-	if err := user.InjectOrgIDIntoHTTPRequest(ctx, req); err != nil {
-		return nil, err
-	}
-
-	return req.WithContext(ctx), nil
+	return nil
 }
 
 // EncodeLabelsSeriesQueryRequest encodes a LabelsSeriesQueryRequest into an http request.
@@ -839,8 +849,8 @@ func (c Codec) EncodeLabelsSeriesQueryRequest(ctx context.Context, req LabelsSer
 		return nil, fmt.Errorf("unknown query result response format '%s'", c.preferredQueryResultResponseFormat)
 	}
 
-	if level, ok := api.ReadConsistencyLevelFromContext(ctx); ok {
-		r.Header.Add(api.ReadConsistencyHeader, level)
+	if err := c.injector.InjectToCarrier(ctx, propagation.HttpHeaderCarrier(r.Header)); err != nil {
+		return nil, err
 	}
 
 	// Propagate allowed HTTP headers.
@@ -873,12 +883,6 @@ func encodeOptions(req *http.Request, o Options) {
 	if o.TotalShards > 0 {
 		req.Header.Set(totalShardsControlHeader, strconv.Itoa(int(o.TotalShards)))
 	}
-	if o.InstantSplitDisabled {
-		req.Header.Set(instantSplitControlHeader, "0")
-	}
-	if o.InstantSplitInterval > 0 {
-		req.Header.Set(instantSplitControlHeader, time.Duration(o.InstantSplitInterval).String())
-	}
 }
 
 // DecodeMetricsQueryResponse decodes a Response from an http response.
@@ -1314,3 +1318,19 @@ func DecorateWithParamName(err error, field string) error {
 	}
 	return apierror.Newf(apierror.TypeBadData, errTmpl, field, err)
 }
+
+type headersContextKeyType int
+
+const headersContextKey headersContextKeyType = iota
+
+func ContextWithHeadersToPropagate(ctx context.Context, headers map[string][]string) context.Context {
+	return context.WithValue(ctx, headersContextKey, headers)
+}
+
+func HeadersToPropagateFromContext(ctx context.Context) map[string][]string {
+	if v := ctx.Value(headersContextKey); v != nil {
+		return v.(map[string][]string)
+	}
+
+	return nil
+}
