@@ -18,13 +18,16 @@ import (
 	"github.com/oklog/ulid/v2"
 	"github.com/pkg/errors"
 	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/model/labels"
 	"github.com/prometheus/prometheus/storage"
 	"github.com/prometheus/prometheus/tsdb"
 	"github.com/prometheus/prometheus/tsdb/chunks"
+	"github.com/prometheus/prometheus/tsdb/index"
 	"go.uber.org/atomic"
 
 	"github.com/grafana/mimir/pkg/ingester/activeseries"
+	"github.com/grafana/mimir/pkg/ingester/lookupplan"
 	"github.com/grafana/mimir/pkg/util/extract"
 	"github.com/grafana/mimir/pkg/util/globalerror"
 	util_math "github.com/grafana/mimir/pkg/util/math"
@@ -86,12 +89,14 @@ var (
 )
 
 type ownedSeriesState struct {
-	ownedSeriesCount int // Number of "owned" series, based on current ring.
-	shardSize        int // Tenant shard size when "owned" series was last updated due to ring or shard size changes. Used to detect shard size changes.
-	localSeriesLimit int // Local series limit when "owned" series was last updated due to ring or shard size changes. Used as a minimum when calculating series limits.
+	ownedSeriesCount           int // Number of "owned" series, based on current ring.
+	ownedTargetInfoSeriesCount int // Number of "owned" target_info series, based on current ring.
+	shardSize                  int // Tenant shard size when "owned" series was last updated due to ring or shard size changes. Used to detect shard size changes.
+	localSeriesLimit           int // Local series limit when "owned" series was last updated due to ring or shard size changes. Used as a minimum when calculating series limits.
 }
 
 type userTSDB struct {
+	cfg            *Config
 	db             *tsdb.DB
 	userID         string
 	activeSeries   *activeseries.ActiveSeries
@@ -142,6 +147,46 @@ type userTSDB struct {
 	ownedTokenRanges ring.TokenRanges
 
 	requiresOwnedSeriesUpdate atomic.String // Non-empty string means that we need to recompute "owned series" for the user. Value will be used in the log message.
+
+	postingsCache *tsdb.PostingsForMatchersCache
+
+	// plannerProvider is optional; if set, it will be used to generate and cache statistics for the user's head block.
+	// Other blocks' stats are immutable and the prometheus TSDB caches them itself.
+	plannerProvider *plannerProvider
+}
+
+// generateHeadStatistics generates statistics for this user's head block.
+func (u *userTSDB) generateHeadStatistics() error {
+	// Open head block
+	head := u.db.Head()
+	indexReader, err := head.Index()
+	if err != nil {
+		return fmt.Errorf("failed to open TSDB head index reader: %w", err)
+	}
+	defer indexReader.Close()
+
+	// Get block metadata
+	blockMeta := head.Meta()
+
+	// Generate statistics
+	u.plannerProvider.generateAndStorePlanner(blockMeta, indexReader)
+	return nil
+}
+
+// getIndexLookupPlannerFunc returns the appropriate index lookup planner function based on configuration.
+// When index lookup planning is enabled, it first checks if a pre-computed planner exists in the repository.
+// If found, it uses the cached planner; otherwise, it falls back to generating statistics on-demand.
+// When disabled, it uses NoopPlanner which performs no optimization.
+// getIndexLookupPlannerFunc returns a cached planner or generates one on-demand.
+// Not all planners are cached after being created, see plannerProvider.getPlanner() for details.
+func (u *userTSDB) getIndexLookupPlannerFunc() tsdb.IndexLookupPlannerFunc {
+	return func(blockMeta tsdb.BlockMeta, indexReader tsdb.IndexReader) index.LookupPlanner {
+		if !u.cfg.BlocksStorageConfig.TSDB.IndexLookupPlanning.Enabled {
+			return lookupplan.NoopPlanner{}
+		}
+
+		return u.plannerProvider.getPlanner(blockMeta, indexReader)
+	}
 }
 
 func (u *userTSDB) Appender(ctx context.Context) storage.Appender {
@@ -324,23 +369,35 @@ func (u *userTSDB) getSeriesCountAndMinLocalLimit() (int, int) {
 	return count, minLocalLimit
 }
 
+// PostCreation implements tsdb.SeriesLifecycleCallback.
 func (u *userTSDB) PostCreation(metric labels.Labels) {
 	u.instanceSeriesCount.Inc()
 
+	metricName, err := extract.MetricNameFromLabels(metric)
+	if err != nil {
+		// This should never happen because it has already been checked in PreCreation().
+		metricName = ""
+	}
+
 	// If series was just created, it must belong to this ingester. (Unless it was created while replaying WAL,
 	// but we will recompute owned series when ingester joins the ring.)
 	u.ownedStateMtx.Lock()
 	u.ownedState.ownedSeriesCount++
+	if metricName == "target_info" {
+		u.ownedState.ownedTargetInfoSeriesCount++
+	}
 	u.ownedStateMtx.Unlock()
 
-	metricName, err := extract.MetricNameFromLabels(metric)
-	if err != nil {
-		// This should never happen because it has already been checked in PreCreation().
+	if metricName == "" {
 		return
 	}
 	u.seriesInMetric.increaseSeriesForMetric(metricName)
+	if u.postingsCache != nil {
+		u.postingsCache.InvalidateMetric(u.userID, metricName)
+	}
 }
 
+// PostDeletion implements tsdb.SeriesLifecycleCallback.
 func (u *userTSDB) PostDeletion(metrics map[chunks.HeadSeriesRef]labels.Labels) {
 	u.instanceSeriesCount.Sub(int64(len(metrics)))
 
@@ -543,10 +600,15 @@ const (
 	recomputeOwnedSeriesMaxSeriesDiff = 1000
 )
 
-func (u *userTSDB) recomputeOwnedSeriesWithComputeFn(shardSize int, reason string, logger log.Logger, compute func() int) (success bool, _ int) {
+func (u *userTSDB) recomputeOwnedSeriesWithComputeFn(shardSize int, reason string, logger log.Logger, compute func() (int, int)) (success bool, _ int) {
 	start := time.Now()
 
-	var ownedSeriesNew, ownedSeriesBefore, shardSizeBefore, localLimitBefore, localLimitNew int
+	var (
+		ownedSeriesNew, ownedSeriesBefore,
+		ownedTargetInfoSeriesNew, ownedTargetInfoSeriesBefore,
+		shardSizeBefore,
+		localLimitBefore, localLimitNew int
+	)
 
 	success = false
 	attempts := 0
@@ -555,11 +617,12 @@ func (u *userTSDB) recomputeOwnedSeriesWithComputeFn(shardSize int, reason strin
 
 		os := u.ownedSeriesState()
 		ownedSeriesBefore = os.ownedSeriesCount
+		ownedTargetInfoSeriesBefore = os.ownedTargetInfoSeriesCount
 		shardSizeBefore = os.shardSize
 		localLimitBefore = os.localSeriesLimit
 
 		localLimitNew = u.limiter.maxSeriesPerUser(u.userID, 0)
-		ownedSeriesNew = compute()
+		ownedSeriesNew, ownedTargetInfoSeriesNew = compute()
 
 		u.ownedStateMtx.Lock()
 
@@ -568,12 +631,14 @@ func (u *userTSDB) recomputeOwnedSeriesWithComputeFn(shardSize int, reason strin
 		// (it may or may not include the new series, we don't know).
 		// In that case, just run the computation again -- if there are more attempts left.
 		seriesDiff := u.ownedState.ownedSeriesCount - ownedSeriesBefore
-		if seriesDiff >= 0 && seriesDiff <= recomputeOwnedSeriesMaxSeriesDiff {
+		targetInfoSeriesDiff := u.ownedState.ownedTargetInfoSeriesCount - ownedTargetInfoSeriesBefore
+		if seriesDiff >= 0 && targetInfoSeriesDiff >= 0 && seriesDiff <= recomputeOwnedSeriesMaxSeriesDiff && targetInfoSeriesDiff <= recomputeOwnedSeriesMaxSeriesDiff {
 			success = true
 		}
 
-		// Even if we run computation again, we can start using our (possibly incorrect) values already.
+		// Even if we run the computation again, we can start using our (possibly incorrect) values already.
 		u.ownedState.ownedSeriesCount = ownedSeriesNew
+		u.ownedState.ownedTargetInfoSeriesCount = ownedTargetInfoSeriesNew
 		u.ownedState.shardSize = shardSize
 		u.ownedState.localSeriesLimit = localLimitNew
 
@@ -591,6 +656,8 @@ func (u *userTSDB) recomputeOwnedSeriesWithComputeFn(shardSize int, reason strin
 		"reason", reason,
 		"ownedSeriesBefore", ownedSeriesBefore,
 		"ownedSeriesNew", ownedSeriesNew,
+		"ownedTargetInfoSeriesBefore", ownedTargetInfoSeriesBefore,
+		"ownedTargetInfoSeriesNew", ownedTargetInfoSeriesNew,
 		"shardSizeBefore", shardSizeBefore,
 		"shardSizeNew", shardSize,
 		"localLimitBefore", localLimitBefore,
@@ -611,14 +678,15 @@ func (u *userTSDB) updateTokenRanges(newTokenRanges []uint32) bool {
 	return !prev.Equal(newTokenRanges)
 }
 
-func (u *userTSDB) computeOwnedSeries() int {
+func (u *userTSDB) computeOwnedSeries() (int, int) {
 	// This can happen if ingester doesn't own this tenant anymore.
 	if len(u.ownedTokenRanges) == 0 {
 		u.activeSeries.Clear()
-		return 0
+		return 0, 0
 	}
 
 	count := 0
+	targetInfoCount := 0
 	idx := u.Head().MustIndex()
 	defer idx.Close()
 
@@ -626,10 +694,14 @@ func (u *userTSDB) computeOwnedSeries() int {
 		for i, sh := range secondaryHashes {
 			if u.ownedTokenRanges.IncludesKey(sh) {
 				count++
+				name, err := idx.LabelValueFor(context.Background(), storage.SeriesRef(refs[i]), model.MetricNameLabel)
+				if err == nil && name == "target_info" {
+					targetInfoCount++
+				}
 			} else {
 				u.activeSeries.Delete(refs[i], idx)
 			}
 		}
 	})
-	return count
+	return count, targetInfoCount
 }
