@@ -13,6 +13,7 @@ import (
 	"os"
 	"path/filepath"
 	goregexp "regexp" //lint:ignore faillint the Prometheus client library requires us to pass a regexp from this package
+	"slices"
 	"strconv"
 	"strings"
 	"time"
@@ -35,7 +36,6 @@ import (
 	"github.com/pkg/errors"
 	"github.com/prometheus/client_golang/prometheus"
 	"github.com/prometheus/client_golang/prometheus/collectors"
-	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/promql"
 	prom_storage "github.com/prometheus/prometheus/storage"
 	"go.uber.org/atomic"
@@ -56,7 +56,6 @@ import (
 	"github.com/grafana/mimir/pkg/flusher"
 	"github.com/grafana/mimir/pkg/frontend"
 	"github.com/grafana/mimir/pkg/frontend/querymiddleware"
-	frontendv1 "github.com/grafana/mimir/pkg/frontend/v1"
 	"github.com/grafana/mimir/pkg/ingester"
 	"github.com/grafana/mimir/pkg/ingester/client"
 	"github.com/grafana/mimir/pkg/mimirpb"
@@ -74,11 +73,13 @@ import (
 	"github.com/grafana/mimir/pkg/storegateway"
 	"github.com/grafana/mimir/pkg/streamingpromql"
 	"github.com/grafana/mimir/pkg/usagestats"
+	"github.com/grafana/mimir/pkg/usagetracker"
 	"github.com/grafana/mimir/pkg/util"
 	"github.com/grafana/mimir/pkg/util/activitytracker"
 	util_log "github.com/grafana/mimir/pkg/util/log"
 	"github.com/grafana/mimir/pkg/util/noauth"
 	"github.com/grafana/mimir/pkg/util/process"
+	"github.com/grafana/mimir/pkg/util/propagation"
 	"github.com/grafana/mimir/pkg/util/validation"
 	"github.com/grafana/mimir/pkg/util/validation/exporter"
 	"github.com/grafana/mimir/pkg/vault"
@@ -114,25 +115,26 @@ type Config struct {
 	PrintConfig                     bool                   `yaml:"-"`
 	ApplicationName                 string                 `yaml:"-"`
 
-	API                   api.Config                      `yaml:"api"`
-	Server                server.Config                   `yaml:"server"`
-	Distributor           distributor.Config              `yaml:"distributor"`
-	Querier               querier.Config                  `yaml:"querier"`
-	IngesterClient        client.Config                   `yaml:"ingester_client"`
-	Ingester              ingester.Config                 `yaml:"ingester"`
-	Flusher               flusher.Config                  `yaml:"flusher"`
-	LimitsConfig          validation.Limits               `yaml:"limits"`
-	Worker                querier_worker.Config           `yaml:"frontend_worker"`
-	Frontend              frontend.CombinedFrontendConfig `yaml:"frontend"`
-	IngestStorage         ingest.Config                   `yaml:"ingest_storage"`
-	BlockBuilder          blockbuilder.Config             `yaml:"block_builder" doc:"hidden"`
-	BlockBuilderScheduler blockbuilderscheduler.Config    `yaml:"block_builder_scheduler" doc:"hidden"`
-	BlocksStorage         tsdb.BlocksStorageConfig        `yaml:"blocks_storage"`
-	Compactor             compactor.Config                `yaml:"compactor"`
-	StoreGateway          storegateway.Config             `yaml:"store_gateway"`
-	TenantFederation      tenantfederation.Config         `yaml:"tenant_federation"`
-	ActivityTracker       activitytracker.Config          `yaml:"activity_tracker"`
-	Vault                 vault.Config                    `yaml:"vault"`
+	API                            api.Config                      `yaml:"api"`
+	Server                         server.Config                   `yaml:"server"`
+	Distributor                    distributor.Config              `yaml:"distributor"`
+	Querier                        querier.Config                  `yaml:"querier"`
+	IngesterClient                 client.Config                   `yaml:"ingester_client"`
+	Ingester                       ingester.Config                 `yaml:"ingester"`
+	Flusher                        flusher.Config                  `yaml:"flusher"`
+	LimitsConfig                   validation.Limits               `yaml:"limits"`
+	Worker                         querier_worker.Config           `yaml:"frontend_worker"`
+	Frontend                       frontend.CombinedFrontendConfig `yaml:"frontend"`
+	IngestStorage                  ingest.Config                   `yaml:"ingest_storage"`
+	BlockBuilder                   blockbuilder.Config             `yaml:"block_builder" doc:"hidden"`
+	BlockBuilderScheduler          blockbuilderscheduler.Config    `yaml:"block_builder_scheduler" doc:"hidden"`
+	BlocksStorage                  tsdb.BlocksStorageConfig        `yaml:"blocks_storage"`
+	Compactor                      compactor.Config                `yaml:"compactor"`
+	StoreGateway                   storegateway.Config             `yaml:"store_gateway"`
+	TenantFederation               tenantfederation.Config         `yaml:"tenant_federation"`
+	ActivityTracker                activitytracker.Config          `yaml:"activity_tracker"`
+	IncludeTenantIDInProfileLabels bool                            `yaml:"include_tenant_id_in_profile_labels" category:"experimental"`
+	Vault                          vault.Config                    `yaml:"vault"`
 
 	Ruler               ruler.Config                               `yaml:"ruler"`
 	RulerStorage        rulestore.Config                           `yaml:"ruler_storage"`
@@ -142,6 +144,7 @@ type Config struct {
 	MemberlistKV        memberlist.KVConfig                        `yaml:"memberlist"`
 	QueryScheduler      scheduler.Config                           `yaml:"query_scheduler"`
 	UsageStats          usagestats.Config                          `yaml:"usage_stats"`
+	UsageTracker        usagetracker.Config                        `yaml:"usage_tracker" doc:"hidden"`
 	ContinuousTest      continuoustest.Config                      `yaml:"-"`
 	OverridesExporter   exporter.Config                            `yaml:"overrides_exporter"`
 
@@ -180,6 +183,7 @@ func (c *Config) RegisterFlags(f *flag.FlagSet, logger log.Logger) {
 	f.StringVar(&c.CostAttributionRegistryPath, "cost-attribution.registry-path", "", "Defines a custom path for the registry. When specified, Mimir exposes cost attribution metrics through this custom path. If not specified, cost attribution metrics aren't exposed.")
 	f.DurationVar(&c.CostAttributionEvictionInterval, "cost-attribution.eviction-interval", 20*time.Minute, "Specifies how often inactive cost attributions for received and discarded sample trackers are evicted from the counter, ensuring they do not contribute to the cost attribution cardinality per user limit. This setting does not apply to active series, which are managed separately.")
 	f.DurationVar(&c.CostAttributionCleanupInterval, "cost-attribution.cleanup-interval", 3*time.Minute, "Time interval at which the cost attribution cleanup process runs, ensuring inactive cost attribution entries are purged.")
+	f.BoolVar(&c.IncludeTenantIDInProfileLabels, "include-tenant-id-in-profile-labels", false, "Include tenant ID in pprof labels for profiling. Currently only supported by the ingester. This can help debug performance issues for specific tenants.")
 
 	c.API.RegisterFlags(f)
 	c.registerServerFlagsWithChangedDefaultValues(f)
@@ -208,6 +212,7 @@ func (c *Config) RegisterFlags(f *flag.FlagSet, logger log.Logger) {
 	c.ActivityTracker.RegisterFlags(f)
 	c.QueryScheduler.RegisterFlags(f, logger)
 	c.UsageStats.RegisterFlags(f)
+	c.UsageTracker.RegisterFlags(f, logger)
 	c.ContinuousTest.RegisterFlags(f)
 	c.OverridesExporter.RegisterFlags(f, logger)
 
@@ -217,9 +222,10 @@ func (c *Config) RegisterFlags(f *flag.FlagSet, logger log.Logger) {
 func (c *Config) CommonConfigInheritance() CommonConfigInheritance {
 	return CommonConfigInheritance{
 		Storage: map[string]*bucket.StorageBackendConfig{
-			"blocks_storage":       &c.BlocksStorage.Bucket.StorageBackendConfig,
-			"ruler_storage":        &c.RulerStorage.StorageBackendConfig,
-			"alertmanager_storage": &c.AlertmanagerStorage.StorageBackendConfig,
+			"blocks_storage":                  &c.BlocksStorage.Bucket.StorageBackendConfig,
+			"ruler_storage":                   &c.RulerStorage.StorageBackendConfig,
+			"alertmanager_storage":            &c.AlertmanagerStorage.StorageBackendConfig,
+			"usage_tracker_snapshots_storage": &c.UsageTracker.SnapshotsStorage.StorageBackendConfig,
 		},
 		ClientClusterValidation: map[string]*clusterutil.ClusterValidationConfig{
 			"ingester_client":                  &c.IngesterClient.GRPCClientConfig.ClusterValidation,
@@ -233,6 +239,7 @@ func (c *Config) CommonConfigInheritance() CommonConfigInheritance {
 			"ruler_client":                     &c.Ruler.ClientTLSConfig.ClusterValidation,
 			"ruler_query_frontend_client":      &c.Ruler.QueryFrontend.GRPCClientConfig.ClusterValidation,
 			"alert_manager_client":             &c.Alertmanager.AlertmanagerClient.GRPCClientConfig.ClusterValidation,
+			"usage_tracker_client":             &c.Distributor.UsageTrackerClient.GRPCClientConfig.ClusterValidation,
 		},
 	}
 }
@@ -270,7 +277,7 @@ func (c *Config) Validate(log log.Logger) error {
 	if err := c.IngestStorage.Validate(); err != nil {
 		return errors.Wrap(err, "invalid ingest storage config")
 	}
-	if c.isAnyModuleEnabled(Ingester, Write, All) {
+	if c.isIngesterEnabled() {
 		if !c.IngestStorage.Enabled && !c.Ingester.PushGrpcMethodEnabled {
 			return errors.New("cannot disable Push gRPC method in ingester, while ingest storage (-ingest-storage.enabled) is not enabled")
 		}
@@ -296,7 +303,7 @@ func (c *Config) Validate(log log.Logger) error {
 		// passing a unique set of per instance flags, e.g. "-ingester.ring.instance-id".
 		// Such a scenario breaks the validation of other modules if those flags aren't also passed to each instance (ref
 		// grafana/mimir#7822). Otherwise, log the fact and move on.
-		if c.isAnyModuleEnabled(Ingester, Write, All) || !errors.Is(err, ingester.ErrSpreadMinimizingValidation) {
+		if c.isIngesterEnabled() || !errors.Is(err, ingester.ErrSpreadMinimizingValidation) {
 			return errors.Wrap(err, "invalid ingester config")
 		}
 		level.Debug(log).Log("msg", "ingester config is invalid; moving on because the \"ingester\" module is not in this process's targets", "err", err.Error())
@@ -322,10 +329,20 @@ func (c *Config) Validate(log log.Logger) error {
 	if err := c.UsageStats.Validate(); err != nil {
 		return errors.Wrap(err, "invalid usage stats config")
 	}
+	if c.isDistributorEnabled() {
+		if err := c.UsageTracker.ValidateForClient(); err != nil {
+			return errors.Wrap(err, "invalid usage-tracker config")
+		}
+	}
+	if c.isUsageTrackerEnabled() {
+		if err := c.UsageTracker.ValidateForUsageTracker(); err != nil {
+			return errors.Wrap(err, "invalid usage-tracker config")
+		}
+	}
 	if err := c.Vault.Validate(); err != nil {
 		return errors.Wrap(err, "invalid vault config")
 	}
-	if c.isAnyModuleEnabled(AlertManager, Backend) {
+	if c.isAlertManagerEnabled() {
 		if err := c.Alertmanager.Validate(); err != nil {
 			return errors.Wrap(err, "invalid alertmanager config")
 		}
@@ -334,41 +351,29 @@ func (c *Config) Validate(log log.Logger) error {
 		return errors.Wrap(err, "invalid overrides-exporter config")
 	}
 
-	if c.Distributor.HATrackerConfig.DeprecatedUpdateTimeout > 0 {
-		c.LimitsConfig.HATrackerUpdateTimeout = model.Duration(c.Distributor.HATrackerConfig.DeprecatedUpdateTimeout)
-		level.Warn(log).Log("msg", "using deprecated config parameter distributor.ha_tracker.ha_tracker_update_timeout; use limits.ha_tracker_update_timeout instead")
-	}
-	if c.Distributor.HATrackerConfig.DeprecatedUpdateTimeoutJitterMax > 0 {
-		c.LimitsConfig.HATrackerUpdateTimeoutJitterMax = model.Duration(c.Distributor.HATrackerConfig.DeprecatedUpdateTimeoutJitterMax)
-		level.Warn(log).Log("msg", "using deprecated config parameter distributor.ha_tracker.ha_tracker_update_timeout_jitter_max; use limits.ha_tracker_update_timeout_jitter_max instead")
-	}
-	if c.Distributor.HATrackerConfig.DeprecatedFailoverTimeout > 0 {
-		c.LimitsConfig.HATrackerFailoverTimeout = model.Duration(c.Distributor.HATrackerConfig.DeprecatedFailoverTimeout)
-		level.Warn(log).Log("msg", "using deprecated config parameter distributor.ha_tracker.ha_tracker_failover_timeout; use limits.ha_tracker_failover_timeout instead")
-	}
 	// validate the default limits
-	if err := c.ValidateLimits(c.LimitsConfig); err != nil {
+	if err := c.ValidateLimits(&c.LimitsConfig); err != nil {
 		return err
 	}
 
 	return nil
 }
 
-// ValidateLimits validates the runtime limits that can be set for each tenant against static configs
-func (c *Config) ValidateLimits(limits validation.Limits) error {
-	if err := c.Querier.ValidateLimits(limits); err != nil {
+// ValidateLimits validates the runtime limits.
+func (c *Config) ValidateLimits(limits *validation.Limits) error {
+	if err := c.Querier.ValidateLimits(*limits); err != nil {
 		return errors.Wrap(err, "invalid limits config for querier")
 	}
-	return nil
+	return limits.Validate()
 }
 
-func (c *Config) isModuleEnabled(m string) bool {
-	return util.StringsContain(c.Target, m)
+func (c *Config) isModuleExplicitlyTargeted(m string) bool {
+	return slices.Contains(c.Target, m)
 }
 
-func (c *Config) isAnyModuleEnabled(modules ...string) bool {
+func (c *Config) isAnyModuleExplicitlyTargeted(modules ...string) bool {
 	for _, m := range modules {
-		if c.isModuleEnabled(m) {
+		if c.isModuleExplicitlyTargeted(m) {
 			return true
 		}
 	}
@@ -376,19 +381,64 @@ func (c *Config) isAnyModuleEnabled(modules ...string) bool {
 	return false
 }
 
+func (c *Config) isIngesterEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, Ingester)
+}
+
+func (c *Config) isAlertManagerEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(AlertManager)
+}
+
+func (c *Config) isRulerEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, Ruler)
+}
+
+func (c *Config) isStoreGatewayEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, StoreGateway)
+}
+
+func (c *Config) isCompactorEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, Compactor)
+}
+
+func (c *Config) isQueryFrontendEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, QueryFrontend)
+}
+
+func (c *Config) isQuerySchedulerEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, QueryScheduler)
+}
+
+func (c *Config) isQuerierEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, Querier)
+}
+
+func (c *Config) isDistributorEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(All, Distributor)
+}
+
+func (c *Config) isUsageTrackerEnabled() bool {
+	return c.isAnyModuleExplicitlyTargeted(UsageTracker)
+}
+
 func (c *Config) validateBucketConfigs() error {
 	errs := multierror.New()
 
 	// Validate alertmanager bucket config.
-	if c.isAnyModuleEnabled(AlertManager, Backend) && c.AlertmanagerStorage.Backend != alertstorelocal.Name {
+	if c.isAlertManagerEnabled() && c.AlertmanagerStorage.Backend != alertstorelocal.Name {
 		errs.Add(errors.Wrap(validateBucketConfig(c.AlertmanagerStorage.Config, c.BlocksStorage.Bucket), "alertmanager storage"))
 	}
 
 	// Validate ruler bucket config.
-	if c.isAnyModuleEnabled(All, Ruler, Backend) && c.RulerStorage.Backend != rulestore.BackendLocal {
+	if c.isRulerEnabled() && c.RulerStorage.Backend != rulestore.BackendLocal {
 		errs.Add(errors.Wrap(validateBucketConfig(c.RulerStorage.Config, c.BlocksStorage.Bucket), "ruler storage"))
 	}
 
+	// Validate usage tracker snapshots bucket config.
+	if c.isUsageTrackerEnabled() && c.UsageTracker.SnapshotsStorage.Backend != bucket.Filesystem {
+		errs.Add(errors.Wrap(validateBucketConfig(c.UsageTracker.SnapshotsStorage, c.BlocksStorage.Bucket), "usage-tracker snapshots storage"))
+	}
+
 	return errs.Err()
 }
 
@@ -441,7 +491,7 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 	var paths []pathConfig
 
 	// Blocks storage (check only for components using it).
-	if c.isAnyModuleEnabled(All, Write, Read, Backend, Ingester, Querier, StoreGateway, Compactor, Ruler) && c.BlocksStorage.Bucket.Backend == bucket.Filesystem {
+	if (c.isIngesterEnabled() || c.isQuerierEnabled() || c.isStoreGatewayEnabled() || c.isCompactorEnabled() || c.isRulerEnabled()) && c.BlocksStorage.Bucket.Backend == bucket.Filesystem {
 		// Add the optional prefix to the path, because that's the actual location where blocks will be stored.
 		paths = append(paths, pathConfig{
 			name:       "blocks storage filesystem directory",
@@ -451,7 +501,7 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 	}
 
 	// Ingester.
-	if c.isAnyModuleEnabled(All, Ingester, Write) {
+	if c.isIngesterEnabled() {
 		paths = append(paths, pathConfig{
 			name:       "tsdb directory",
 			cfgValue:   c.BlocksStorage.TSDB.Dir,
@@ -460,7 +510,7 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 	}
 
 	// Store-gateway.
-	if c.isAnyModuleEnabled(All, StoreGateway, Backend) {
+	if c.isStoreGatewayEnabled() {
 		paths = append(paths, pathConfig{
 			name:       "bucket store sync directory",
 			cfgValue:   c.BlocksStorage.BucketStore.SyncDir,
@@ -469,7 +519,7 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 	}
 
 	// Compactor.
-	if c.isAnyModuleEnabled(All, Compactor, Backend) {
+	if c.isCompactorEnabled() {
 		paths = append(paths, pathConfig{
 			name:       "compactor data directory",
 			cfgValue:   c.Compactor.DataDir,
@@ -478,7 +528,7 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 	}
 
 	// Ruler.
-	if c.isAnyModuleEnabled(All, Ruler, Backend) {
+	if c.isRulerEnabled() {
 		paths = append(paths, pathConfig{
 			name:       "ruler data directory",
 			cfgValue:   c.Ruler.RulePath,
@@ -503,7 +553,7 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 	}
 
 	// Alertmanager.
-	if c.isAnyModuleEnabled(AlertManager, Backend) {
+	if c.isAlertManagerEnabled() {
 		paths = append(paths, pathConfig{
 			name:       "alertmanager data directory",
 			cfgValue:   c.Alertmanager.DataDir,
@@ -530,6 +580,17 @@ func (c *Config) validateFilesystemPaths(logger log.Logger) error {
 		}
 	}
 
+	// Usage-tracker.
+	if c.isUsageTrackerEnabled() {
+		if c.UsageTracker.SnapshotsStorage.Backend == bucket.Filesystem {
+			paths = append(paths, pathConfig{
+				name:       "usage tracker snapshot storage filesystem directory",
+				cfgValue:   c.UsageTracker.SnapshotsStorage.Filesystem.Directory,
+				checkValue: filepath.Join(c.UsageTracker.SnapshotsStorage.Filesystem.Directory, usagetracker.SnapshotsStoragePrefix),
+			})
+		}
+	}
+
 	// Convert all check paths to absolute clean paths.
 	for idx, path := range paths {
 		abs, err := filepath.Abs(path.checkValue)
@@ -777,6 +838,7 @@ type Mimir struct {
 
 	API                              *api.API
 	Server                           *server.Server
+	ServerMetrics                    *server.Metrics
 	IngesterRing                     *ring.Ring
 	IngesterPartitionRingWatcher     *ring.PartitionRingWatcher
 	IngesterPartitionInstanceRing    *ring.PartitionInstanceRing
@@ -786,14 +848,14 @@ type Mimir struct {
 	Distributor                      *distributor.Distributor
 	Ingester                         *ingester.Ingester
 	Flusher                          *flusher.Flusher
-	FrontendV1                       *frontendv1.Frontend
 	RuntimeConfig                    *runtimeconfig.Manager
 	QuerierQueryable                 prom_storage.SampleAndChunkQueryable
 	ExemplarQueryable                prom_storage.ExemplarQueryable
 	AdditionalStorageQueryables      []querier.TimeRangeQueryable
 	MetadataSupplier                 querier.MetadataSupplier
 	QuerierEngine                    promql.QueryEngine
-	QueryPlanner                     *streamingpromql.QueryPlanner
+	QuerierStreamingEngine           *streamingpromql.Engine // The MQE instance in QuerierEngine (without fallback wrapper), or nil if MQE is disabled.
+	QueryFrontendStreamingEngine     *streamingpromql.Engine // The MQE instance used by the query-frontend (without fallback wrapper), or nil if MQE is disabled.
 	QueryFrontendTripperware         querymiddleware.Tripperware
 	QueryFrontendTopicOffsetsReaders map[string]*ingest.TopicOffsetsReader
 	QueryFrontendCodec               querymiddleware.Codec
@@ -806,11 +868,29 @@ type Mimir struct {
 	ActivityTracker                  *activitytracker.ActivityTracker
 	Vault                            *vault.Vault
 	UsageStatsReporter               *usagestats.Reporter
+	UsageTracker                     *usagetracker.UsageTracker
+	UsageTrackerPartitionRing        *ring.MultiPartitionInstanceRing
+	UsageTrackerInstanceRing         *ring.Ring
 	BlockBuilder                     *blockbuilder.BlockBuilder
 	BlockBuilderScheduler            *blockbuilderscheduler.BlockBuilderScheduler
 	ContinuousTestManager            *continuoustest.Manager
 	BuildInfoHandler                 http.Handler
 	CostAttributionManager           *costattribution.Manager
+
+	// Extractors are used by queriers to extract HTTP headers / metadata from incoming requests.
+	// We use an abstraction here to support both httpgrpc requests and Protobuf requests.
+	Extractors []propagation.Extractor
+
+	// Injectors are used by query-frontends to inject HTTP headers / metadata into outgoing requests to queriers.
+	// We use an abstraction here to support both httpgrpc requests and Protobuf requests.
+	Injectors []propagation.Injector
+
+	QueryFrontendQueryPlanner *streamingpromql.QueryPlanner
+	// The separate planner for queriers is a temporary thing until all query planning is happening solely in query-frontends,
+	// and support for sending queries directly to queriers via the Prometheus HTTP API endpoints is removed.
+	// Until then, we need separate instances as the remote execution optimisation pass must only be applied in query-frontends,
+	// including when running in monolithic and read/write modes.
+	QuerierQueryPlanner *streamingpromql.QueryPlanner
 }
 
 // New makes a new Mimir.
@@ -830,6 +910,13 @@ func New(cfg Config, reg prometheus.Registerer) (*Mimir, error) {
 	cfg.Server.GRPCMiddleware = append(cfg.Server.GRPCMiddleware, querierapi.ReadConsistencyServerUnaryInterceptor)
 	cfg.Server.GRPCStreamMiddleware = append(cfg.Server.GRPCStreamMiddleware, querierapi.ReadConsistencyServerStreamInterceptor)
 
+	if cfg.isIngesterEnabled() {
+		// Propagate only-replica markers for ingester queries
+		cfg.Server.GRPCTapHandles = append(cfg.Server.GRPCTapHandles, client.OnlyReplicaTapHandle)
+		cfg.Server.GRPCMiddleware = append(cfg.Server.GRPCMiddleware, client.OnlyReplicaServerUnaryInterceptor)
+		cfg.Server.GRPCStreamMiddleware = append(cfg.Server.GRPCStreamMiddleware, client.OnlyReplicaServerStreamInterceptor)
+	}
+
 	cfg.API.HTTPAuthMiddleware = noauth.SetupAuthMiddleware(
 		&cfg.Server,
 		cfg.MultitenancyEnabled,
@@ -1055,11 +1142,9 @@ func (t *Mimir) readyHandler(sm *services.Manager, shutdownRequested *atomic.Boo
 			}
 		}
 
-		// Query Frontend has a special check that makes sure that a querier is attached before it signals
-		// itself as ready
-		if t.FrontendV1 != nil {
-			if err := t.FrontendV1.CheckReady(r.Context()); err != nil {
-				http.Error(w, "Query Frontend not ready: "+err.Error(), http.StatusServiceUnavailable)
+		if t.UsageTracker != nil {
+			if err := t.UsageTracker.CheckReady(r.Context()); err != nil {
+				http.Error(w, "Usage Tracker not ready: "+err.Error(), http.StatusServiceUnavailable)
 				return
 			}
 		}
