@@ -668,7 +668,7 @@ func (a *API) CreateRuleGroup(w http.ResponseWriter, req *http.Request) {
 		return
 	}
 
-	errs := a.ruler.manager.ValidateRuleGroup(rg, node)
+	errs := a.ruler.manager.ValidateRuleGroup(userID, rg, node)
 	if len(errs) > 0 {
 		e := []string{}
 		for _, err := range errs {
