@@ -199,10 +199,14 @@ func NewGroupedVectorVectorBinaryOperation(
 // (The alternative would be to compute the entire result here in SeriesMetadata and only return the series that
 // contain points, but that would mean we'd need to hold the entire result in memory at once, which we want to
 // avoid.)
-func (g *GroupedVectorVectorBinaryOperation) SeriesMetadata(ctx context.Context) ([]types.SeriesMetadata, error) {
-	if canProduceAnySeries, err := g.loadSeriesMetadata(ctx); err != nil {
+func (g *GroupedVectorVectorBinaryOperation) SeriesMetadata(ctx context.Context, matchers types.Matchers) ([]types.SeriesMetadata, error) {
+	if canProduceAnySeries, err := g.loadSeriesMetadata(ctx, matchers); err != nil {
 		return nil, err
 	} else if !canProduceAnySeries {
+		if err := g.Finalize(ctx); err != nil {
+			return nil, err
+		}
+
 		g.Close()
 		return nil, nil
 	}
@@ -216,6 +220,11 @@ func (g *GroupedVectorVectorBinaryOperation) SeriesMetadata(ctx context.Context)
 		types.SeriesMetadataSlicePool.Put(&allMetadata, g.MemoryConsumptionTracker)
 		types.BoolSlicePool.Put(&oneSideSeriesUsed, g.MemoryConsumptionTracker)
 		types.BoolSlicePool.Put(&manySideSeriesUsed, g.MemoryConsumptionTracker)
+
+		if err := g.Finalize(ctx); err != nil {
+			return nil, err
+		}
+
 		g.Close()
 		return nil, nil
 	}
@@ -232,12 +241,12 @@ func (g *GroupedVectorVectorBinaryOperation) SeriesMetadata(ctx context.Context)
 // loadSeriesMetadata loads series metadata from both sides of this operation.
 // It returns false if one side returned no series and that means there is no way for this operation to return any series.
 // (eg. if doing A + B and either A or B have no series, then there is no way for this operation to produce any series)
-func (g *GroupedVectorVectorBinaryOperation) loadSeriesMetadata(ctx context.Context) (bool, error) {
+func (g *GroupedVectorVectorBinaryOperation) loadSeriesMetadata(ctx context.Context, matchers types.Matchers) (bool, error) {
 	// We retain the series labels for later so we can use them to generate error messages.
 	// We'll return them to the pool in Close().
 
 	var err error
-	g.oneSideMetadata, err = g.oneSide.SeriesMetadata(ctx)
+	g.oneSideMetadata, err = g.oneSide.SeriesMetadata(ctx, matchers)
 	if err != nil {
 		return false, err
 	}
@@ -247,7 +256,7 @@ func (g *GroupedVectorVectorBinaryOperation) loadSeriesMetadata(ctx context.Cont
 		return false, nil
 	}
 
-	g.manySideMetadata, err = g.manySide.SeriesMetadata(ctx)
+	g.manySideMetadata, err = g.manySide.SeriesMetadata(ctx, matchers)
 	if err != nil {
 		return false, err
 	}
@@ -759,13 +768,21 @@ func (g *GroupedVectorVectorBinaryOperation) ExpressionPosition() posrange.Posit
 }
 
 func (g *GroupedVectorVectorBinaryOperation) Prepare(ctx context.Context, params *types.PrepareParams) error {
-	err := g.Left.Prepare(ctx, params)
-	if err != nil {
+	if err := g.Left.Prepare(ctx, params); err != nil {
 		return err
 	}
+
 	return g.Right.Prepare(ctx, params)
 }
 
+func (g *GroupedVectorVectorBinaryOperation) Finalize(ctx context.Context) error {
+	if err := g.Left.Finalize(ctx); err != nil {
+		return err
+	}
+
+	return g.Right.Finalize(ctx)
+}
+
 func (g *GroupedVectorVectorBinaryOperation) Close() {
 	g.Left.Close()
 	g.Right.Close()
