@@ -6,14 +6,14 @@
 package validation
 
 import (
+	"slices"
+
 	"github.com/grafana/dskit/flagext"
 	"github.com/pkg/errors"
-
-	"github.com/grafana/mimir/pkg/util"
 )
 
 func validateIntegrationLimit(k string, _ float64) error {
-	if !util.StringsContain(allowedIntegrationNames, k) {
+	if !slices.Contains(allowedIntegrationNames, k) {
 		return errors.Errorf("unknown integration name: %s", k)
 	}
 	return nil
@@ -21,7 +21,7 @@ func validateIntegrationLimit(k string, _ float64) error {
 
 // allowedIntegrationNames is a list of all the integrations that can be rate limited.
 var allowedIntegrationNames = []string{
-	"webhook", "email", "pagerduty", "opsgenie", "wechat", "slack", "victorops", "pushover", "sns", "webex", "telegram", "discord", "msteams",
+	"webhook", "email", "pagerduty", "opsgenie", "wechat", "slack", "victorops", "pushover", "sns", "webex", "telegram", "discord", "msteams", "msteamsv2",
 }
 
 // NotificationRateLimitMap returns a map that can be used as a flag for setting notification rate limits.
